import pytest

from hintedi import HinteDI, InjectionException


@HinteDI.singleton
class SingletonDependency:
    pass


@HinteDI.instance
class InstanceDependency:
    pass


class TestInjector(HinteDI):  # isolate test cases here to allow resetting constructor
    pass


class TestHinteDI:

    def test_singleton_when_used_as_dependency_then_same_instance_injected(self):
        class Injected:

            @HinteDI.inject
            def __init__(self, dependency: SingletonDependency):
                self.dependency = dependency

        first_injected = Injected()
        second_injected = Injected()
        assert type(first_injected.dependency) is SingletonDependency
        assert first_injected.dependency is second_injected.dependency

    def test_instance_when_used_as_dependency_then_different_instance_injected(self):
        class Injected:

            @HinteDI.inject
            def __init__(self, dependency: InstanceDependency):
                self.dependency = dependency

        first_injected = Injected()
        second_injected = Injected()
        assert type(first_injected.dependency) is InstanceDependency
        assert type(second_injected.dependency) is InstanceDependency
        assert first_injected.dependency is not second_injected.dependency

    def test_inject_when_injected_with_all_dependencies_provided_then_no_dependency_injected(self):
        @HinteDI.instance
        class Injectable:
            def __init__(self, value: int = 1):
                self.value = value

        def provided_test(first: Injectable, second: Injectable):
            return first.value != 1 and second.value != 1

        assert provided_test(Injectable(2), Injectable(3))

    def test_inject_given_function_with_default_arguments_when_injected_then_default_arguments_not_injected(self):
        @HinteDI.instance
        class Injectable:
            def __init__(self, value: int = 1):
                self.value = value

        @HinteDI.inject
        def default_test(argument: Injectable, another: Injectable = Injectable(2)):
            return another.value > argument.value

        assert default_test()

    def test_inject_given_used_with_method_when_injected_then_self_argument_not_injected(self):
        @HinteDI.singleton
        class Injectable:
            def __init__(self, value: int = 1):
                self.value = value

        class Dependent:

            @HinteDI.inject
            def __init__(self, injectable: Injectable):
                self.injectable = injectable

        assert Injectable().value == Dependent().injectable.value

    def test_inject_given_used_with_class_method_when_injected_then_cls_argument_not_injected(self):
        @HinteDI.singleton
        class Injectable:
            def __init__(self, value: int = 1):
                self.value = value

        class Dependent:

            @classmethod
            @HinteDI.inject
            def method(cls, injectable: Injectable):
                return injectable.value

        assert Injectable().value == Dependent.method()

    def test_inject_given_args_possible_when_injected_then_args_not_injected(self):
        @HinteDI.singleton
        class Injectable:

            def __init__(self, value: bool = True):
                self.value = value

        @HinteDI.inject
        def args_test(argument: Injectable, *args):
            return argument.value and not args

        assert args_test()

    def test_inject_given_kwargs_possible_when_injected_then_kwargs_not_injected(self):
        @HinteDI.singleton
        class Injectable:

            def __init__(self, value: bool = True):
                self.value = value

        @HinteDI.inject
        def args_test(argument: Injectable, **kwargs):
            return argument.value and not kwargs

        assert args_test()

    def test_inject_given_kwargs_present_when_injected_then_dependency_still_injected(self):
        @HinteDI.singleton
        class Injectable:

            def __init__(self, value: bool = True):
                self.value = value

        @HinteDI.inject
        def args_test(argument: Injectable, **kwargs):
            return argument.value and kwargs

        assert args_test(extra="argument")

    def test_inject_when_injecting_method_with_default_args_and_kwargs_then_only_dependencies_injected(self):
        @HinteDI.singleton
        class Injectable:

            def __init__(self, value: bool = True):
                self.value = value

        class Dependent:

            @HinteDI.inject
            def __init__(self, injectable: Injectable, default: bool = True, *args, **kwargs):
                self.injectable = injectable
                self.default = default
                self.args = args
                self.kwargs = kwargs

            def __bool__(self):
                return self.injectable and self.default and not self.args and not self.kwargs

        assert Dependent()

    def test_inject_given_dependency_with_no_resolution_then_error_raised_with_missing_dependency(self):
        class Dependency:
            pass

        class DependentClass:
            @TestInjector.inject
            def __init__(self, dependency: Dependency):
                self.dependency = dependency
        
        expected_exception = InjectionException(f"Could not inject argument dependency because type "
                                                f"'{Dependency.__name__}' is not registered as a dependency.")

        with pytest.raises(InjectionException) as actual_exception:
            DependentClass()
        assert str(expected_exception) == str(actual_exception.value)

    def test_inject_given_dependency_with_no_annotation_then_error_raised_with_missing_annotation(self):
        class DependentClass:
            @TestInjector.inject
            def __init__(self, dependency):
                self.dependency = dependency

        expected_exception = InjectionException(f"Could not inject argument dependency because it doesn't have a type "
                                                f"annotation.")

        with pytest.raises(InjectionException) as actual_exception:
            DependentClass()
        assert str(expected_exception) == str(actual_exception.value)

    @pytest.mark.parametrize("dependency_type", [HinteDI.singleton, HinteDI.instance, HinteDI.abstract_base])
    def test_singleton_given_dependency_of_same_type_then_error_raised_with_duplicate_dependency(self, dependency_type):
        @dependency_type
        class DuplicatedDependency:
            pass

        expected_exception = InjectionException(f"Could not register dependency {DuplicatedDependency.__name__} as a "
                                                f"dependency with identical name already exists.")

        with pytest.raises(InjectionException) as actual_exception:
            HinteDI.singleton(DuplicatedDependency)
        assert str(expected_exception) == str(actual_exception.value)

    @pytest.mark.parametrize("dependency_type", [HinteDI.singleton, HinteDI.instance, HinteDI.abstract_base])
    def test_instance_given_dependency_of_same_type_then_error_raised_with_duplicate_dependency(self, dependency_type):
        @dependency_type
        class DuplicatedDependency:
            pass

        expected_exception = InjectionException(f"Could not register dependency {DuplicatedDependency.__name__} as a "
                                                f"dependency with identical name already exists.")

        with pytest.raises(InjectionException) as actual_exception:
            HinteDI.instance(DuplicatedDependency)
        assert str(expected_exception) == str(actual_exception.value)



__version__ = "0.3.0"
__author__ = "Eetu Asikainen"

from inspect import signature, Parameter
from typing import Any, Callable, Union, Dict


def _is_injectable_argument(arg):
    return arg.kind in (Parameter.POSITIONAL_OR_KEYWORD, Parameter.POSITIONAL_ONLY) and arg.default is Parameter.empty


def _is_method_target_argument(arg):
    return arg.name in ("self", "cls")


class InstanceSentinel:
    """A sentinel class used for differentiating instance dependencies from singleton dependencies in dependency dict.

    This is exposed to enable typechecking against this class.
    """
    pass


class InjectionException(Exception):
    """An exception thrown by dependency injector.

    This is exposed to enable catching it explicitly if needed. It is not advisable to use this class to construct your
    own exceptions.

    :param message: The error message
    :type message: str
    """

    def __init__(self, message: str):
        full_message = f"An error occurred during HinteDI dependency injection: {message}\n If you think this is a " \
                       f"bug, please file a bug report in the HinteDI repository issue tracker in GitHub " \
                       f"(https://github.com/EddieTheCubeHead/HinteDI/issues) using the \"Bug report\" issue template."
        super().__init__(full_message)


class ImplementationFactory:
    """A class returned when resolving an abstract base class with no default implementations. Contains the
    ''ImplementationFactory.resolve_from_key'' -method for resolving a concrete implementation from a given key."""

    def __init__(self, resolver: Callable[[str], Any], base: type):
        self._resolver = resolver
        self._base = base

    def resolve_from_key(self, key: str) -> Any:
        """A method for resolving a concrete implementation for the abstract base class an implementation factory
        represents.

        :param key: The key used for resolving the implementation
        :type key: str"""
        return self._resolver(key)


class DefaultImplementationSentinel:
    """A sentinel class used for identifying default implementations in dependency dict.

    This is exposed to enable typechecking against this class.
    """
    pass


_implementation_dict = Dict[Union[str, DefaultImplementationSentinel], Union[type, ImplementationFactory]]


class HinteDI:
    """The dependency injector class.

    This class is meant to be used through the class method decorators ``HinteDI.singleton``, ``HinteDI.instance``
    and ``HinteDI.inject`` for normal dependency injection and the class method decorators ``HinteDI.abstract_base``,
    ``HinteDI.singleton_implementation`` and ``HinteDI.instance_implementation`` for factory-type abstract base class
    based injection.

    This is provided as a class instead of as a collection of methods to enable subclassing. By subclassing this class
    you can have multiple independent injector implementations in your program. The class also exposes the dependencies
    dict that stores all registered dependencies. This can be manipulated for example for testing purposes. For
    instance dependencies, the dependency type is mapped into an ``InstanceSentinel`` instance, for uninitialized
    singletons into ``None``, for initialized singletons, the singleton instance and for Abstract base classes, a
    dict storing the key-implementation mappings. In the key-implementation mapping the default implementation is stored
    with a special key ``HinteDI.default_implementation``.

    :cvar dependencies: A dict mapping dependency types into the dependency.
    :type dependencies: dict[type, Any]
    :cvar default_implementation: A sentinel object marking default implementations for abstract dependencies
    :type default_implementation: DefaultImplementationSentinel
    """

    dependencies: dict[type, Any] = {}
    default_implementation: DefaultImplementationSentinel = DefaultImplementationSentinel()

    @classmethod
    def singleton(cls, dependency_class: type) -> type:
        """A class method meant to be used as a class decorator to mark the class as a singleton dependency (reuse the
        same instance for all injections).

        Usage::

            @HinteDI.singleton
            class SingletonDependency:
                ...

        :param dependency_class: The class to be decorated
        :type dependency_class: type
        """
        cls._add_dependency(dependency_class, None)
        return dependency_class

    @classmethod
    def instance(cls, dependency_class: type) -> type:
        """A class method meant to be used as a class decorator to mark the class as an instance dependency (create a
        new instance for all injections).

        Usage::

            @HinteDI.instance
            class InstanceDependency:
                ...

        :param dependency_class: The class to be decorated
        :type dependency_class: type
        """
        cls._add_dependency(dependency_class, InstanceSentinel())
        return dependency_class

    @classmethod
    def abstract_base(cls, dependency_class: type) -> type:
        """A class method meant to be used as a class decorator to mark the class as an abstract dependency.

        Use class decorators ``HinteDI.singleton_implementation`` and ``HinteDI.instance_implementation`` to register
        key-based implementations for the class.

        Usage::

            @HinteDI.abstract_base
            class AbstractDependency:
                ...

        :param dependency_class: The class to be decorated
        :type dependency_class: type
        """
        base_dict = {
            cls.default_implementation: ImplementationFactory(cls._resolve_from_key(dependency_class), dependency_class)
        }
        cls._add_dependency(dependency_class, base_dict)
        return dependency_class

    @classmethod
    def singleton_implementation(cls, *, base: type, key: str, is_default: bool = False) -> Callable[[type], type]:
        """A class method meant to be used as a class decorator to mark the class as a concrete singleton-based
        implementation of an abstract dependency.

        See ``HinteDI.abstract_base`` for registering the abstract base class and ``HinteDI.singleton`` for a more
        thorough explanation of the singleton patter in HinteDI

        Usage::

            # Create the abstract dependency first
            @HinteDI.abstract_base
            class AbstractDependency:
                ...

            # Note that actually inheriting the abstract base is not required for HinteDI to work
            @HinteDI.singleton_implementation(base = AbstractDependency, key = "concrete")
            class ConcreteDependency(AbstractDependency):
                ...

        The abstract class injection pattern in HinteDI will return a ``HinteDI.ImplementationFactory`` instance that
        has the ``HinteDI.ImplementationFactory.resolve_from_key`` method that can be used to resolve the concrete instance
        based on a given key. You can also specify a default implementation with the optional ''is_default'' argument for
        the implementation decorators. If a default implementation is present, HinteDI will inject the default
        implementation when asked to inject the abstract base class and add the ''FromKey'' method to the returned
        object enabling you to resolve the dependency to some other implementation if needed.

        Creating a default implementation::

            # Even though the key here is 'default', any key works for this
            @HinteDI.singleton_implementation(base = AbstractDependency, key = "default", is_default = True)
            class DefaultImplementation(AbstractDependency):
                ...

        :param base: The abstract base class the created concrete class will implement
        :type base: type
        :param key: The key that can be used to resolve the abstract base class into the created implementation
        :type key: str
        :param is_default: An optional flag that marks the implementation as the default implementation, default=False
        :type is_default: bool
        """
        def wrapper(dependency_class: type):
            return cls._create_implementation(base, key, is_default, dependency_class, None)

        return wrapper

    @classmethod
    def instance_implementation(cls, *, base: type, key: str, is_default: bool = False) -> Callable[[type], type]:
        """A class method meant to be used as a class decorator to mark the class as a concrete instance-based
        implementation of an abstract dependency.

        See ``HinteDI.abstract_base`` for registering the abstract base class

        Usage::

            # Create the abstract dependency first
            @HinteDI.abstract_base
            class AbstractDependency:
                ...

            # Note that actually inheriting the abstract base is not required for HinteDI to work
            @HinteDI.instance_implementation(base = AbstractDependency, key = "concrete")
            class ConcreteDependency(AbstractDependency):
                ...

        The abstract class injection pattern in HinteDI will return a ``HinteDI.ImplementationFactory`` instance that
        has the ``HinteDI.ImplementationFactory.resolve_from_key`` method that can be used to resolve the concrete instance
        based on a given key. You can also specify a default implementation with the optional ''is_default'' argument for
        the implementation decorators. If a default implementation is present, HinteDI will inject the default
        implementation when asked to inject the abstract base class and add the ''FromKey'' method to the returned
        object enabling you to resolve the dependency to some other implementation if needed.

        Creating a default implementation::

            # Even though the key here is 'default', any key works for this
            @HinteDI.instance_implementation(base = AbstractDependency, key = "default", is_default = True)
            class DefaultImplementation(AbstractDependency):
                ...

        :param base: The abstract base class the created concrete class will implement
        :type base: type
        :param key: The key that can be used to resolve the abstract base class into the created implementation
        :type key: str
        :param is_default: An optional flag that marks the implementation as the default implementation, default=False
        :type is_default: bool
        """
        def wrapper(dependency_class: type):
            return cls._create_implementation(base, key, is_default, dependency_class, InstanceSentinel())

        return wrapper

    @classmethod
    def _create_implementation(cls, base: type, key: str, is_default: bool, dependency_class: type,
                               default_value: None | InstanceSentinel) -> type:
        cls._assert_base_present(base, dependency_class)
        cls._create_key(base, key, dependency_class, is_default)
        cls._add_dependency(dependency_class, default_value)
        return dependency_class

    @classmethod
    def _assert_base_present(cls, base: type, implementation: type):
        if base not in cls.dependencies:
            raise InjectionException(f"Could not register a complete implementation {implementation.__name__} for "
                                     f"class {base.__name__} because {base.__name__} is not registered as abstract "
                                     f"base dependency.")

    @classmethod
    def _create_key(cls, base: type, key: str, dependency_class: type, is_default: bool = True):
        if is_default and type(cls.dependencies[base][cls.default_implementation]) != ImplementationFactory:
            raise InjectionException(f"Could not create default implementation {dependency_class.__name__} for base "
                                     f"{base.__name__} because default implementation already exists as class "
                                     f"{cls.dependencies[base][cls.default_implementation].__name__}.")
        if key in cls.dependencies[base]:
            raise InjectionException(f"Could not create implementation {dependency_class.__name__} for base "
                                     f"{base.__name__} because implementation with the key '{key}' already exists as "
                                     f"class {cls.dependencies[base][key].__name__}.")
        if is_default and hasattr(dependency_class, "resolve_from_key"):
            raise InjectionException(f"Could not create default implementation {dependency_class.__name__} for base "
                                     f"{base.__name__} because default implementations cannot have an attribute named "
                                     f"'resolve_from_key'.")
        cls.dependencies[base][key] = dependency_class
        if is_default:
            cls.dependencies[base][cls.default_implementation] = dependency_class

    @classmethod
    def _add_dependency(cls, dependency_class: type, default_value: Union[None, InstanceSentinel, _implementation_dict]):
        if dependency_class in cls.dependencies:
            raise InjectionException(f"Could not register dependency {dependency_class.__name__} as a dependency with"
                                     f" identical name already exists.")
        cls.dependencies[dependency_class] = default_value

    @classmethod
    def inject(cls, func: Callable):
        """A class method meant to be used as a function decorator to mark the method as requiring injection. Only
        positional or positional + keyword arguments will be injected, keyword-only arguments will not be injected.
        Arguments already provided including default values will be ignored, as will all arguments named self or cls.

        Usage in an instance method::

            class DependentClass:

                @HinteDI.inject
                def __init__(self, dependency: Dependency)  # self is ignored automatically
                    ...

        Usage in a class method::

            class DependentClass:

                @classmethod  # Ensure the classmethod decorator comes first
                @HinteDI.inject
                def perform_class_method(cls, dependency: Dependency)  # cls is ignored automatically
                    ...

        Usage in a function::

            @HinteDI.inject
            def perform_function(dependency: Dependency)
                ...

        If HinteDI is asked to inject an abstract dependency it will return either an ``HinteDI.ImplementationFactory``
        if the dependency has no default implementation, or the default implementation if the dependency has one. If
        a default implementation is returned, HinteDI will add the ''resolve_from_key'' method to the object enabling you to
        resolve it to another implementation if needed.

        Usage with an abstract dependency with no default implementation::

            @HinteDI.inject
            def perform_function(dependency: AbstractDependency):
                concrete_dependency = dependency.resolve_from_key("key")
                ...

        Usage with an abstract dependency with a default implementation::

            @HinteDI.inject
            def perform_function(dependency: AbstractDependency):
                # You can use the returned dependency right away...
                dependency.do_stuff()

                # ...Or you can resolve it into a dependency of another type
                resolved_dependency = dependency.resolve_from_key("key")

        :param func: The function requiring dependency injection
        :type func: Callable
        """
        def injection_wrapper(*args, **kwargs):
            return cls._inject_needed(func, *args, **kwargs)

        return injection_wrapper

    @classmethod
    def _inject_needed(cls, func: Callable, *args, **kwargs):
        func_args = signature(func)
        injected = []
        for arg in list(func_args.parameters.values())[len(args):]:
            if _is_injectable_argument(arg) and not _is_method_target_argument(arg):
                injected += [cls._inject_arg(arg)]
        return func(*args, *injected, **kwargs)

    @classmethod
    def _inject_arg(cls, arg: Parameter):
        if arg.annotation == Parameter.empty:
            raise (InjectionException(f"Could not inject argument {arg} because it doesn't have a type annotation."))
        real_key = cls._is_dependency_present(arg.annotation)
        if real_key:
            return cls._get_dependency(real_key)
        raise (InjectionException(f"Could not inject argument {arg.name} because type '{arg.annotation.__name__}' is "
                                  f"not registered as a dependency."))

    @classmethod
    def _is_dependency_present(cls, annotation: type | str):
        if annotation in cls.dependencies:
            return annotation
        for dependency in cls.dependencies:
            # If inner class shenanigans are happening annotation might arrive as a string, this should be parsed to
            # preserve the functionality of the injector for inner classes as well
            if annotation == str(dependency).split(".")[-1][:-2] or annotation == str(dependency):
                return dependency
        return None

    @classmethod
    def _get_dependency(cls, annotation: type):
        if type(cls.dependencies[annotation]) is InstanceSentinel:
            return annotation()
        if type(cls.dependencies[annotation]) is dict:
            return cls._get_default_abstract_dependency(annotation)
        if cls.dependencies[annotation] is None:
            cls.dependencies[annotation] = annotation()
        return cls.dependencies[annotation]

    @classmethod
    def _get_default_abstract_dependency(cls, abstract_base: type, key: str = "default"):
        cls._assert_implemented(abstract_base)
        key = cls._assert_valid_key_or_default(abstract_base, key)
        if type(cls.dependencies[abstract_base][key]) is ImplementationFactory:
            return cls.dependencies[abstract_base][key]
        real_resolved_type = cls._is_dependency_present(cls.dependencies[abstract_base][key])
        dependency = cls._get_dependency(real_resolved_type)
        if not hasattr(dependency, "resolve_from_key"):
            dependency.resolve_from_key = cls._resolve_from_key(abstract_base)
        return dependency

    @classmethod
    def _assert_implemented(cls, abstract_base: type):
        if len(cls.dependencies[abstract_base]) == 1:
            raise InjectionException(f"Could not inject an implementation of abstract base class "
                                     f"{abstract_base.__name__} because no implementations exist.")

    @classmethod
    def _assert_valid_key_or_default(cls, abstract_base: type, key: str) -> str:
        if key not in cls.dependencies[abstract_base]:
            key = cls.default_implementation
        return key

    @classmethod
    def _resolve_from_key(cls, base: type) -> Callable[[str], Any]:
        def wrapper(key: str) -> Any:
            if key in cls.dependencies[base]:
                return cls._get_dependency(cls.dependencies[base][key])
            raise InjectionException(f"Could not resolve key '{key}' for class {base.__name__}.")

        return wrapper



import pytest

from hintedi import HinteDI, InjectionException, ImplementationFactory


@HinteDI.abstract_base
class AbstractBaseDependency:
    pass


@HinteDI.singleton_implementation(base=AbstractBaseDependency, key="default", is_default=True)
class DefaultConcreteDependency(AbstractBaseDependency):
    pass


@HinteDI.instance_implementation(base=AbstractBaseDependency, key="instance")
class InstanceConcreteDependency(AbstractBaseDependency):
    pass


class TestInjector(HinteDI):  # isolate test cases again to allow resetting constructor
    pass


class TestKeyBasedInjection:

    def test_abstract_base_dependency_when_not_resolved_with_key_then_default_used(self):
        class Injected:

            @HinteDI.inject
            def __init__(self, dependency: AbstractBaseDependency):
                self.dependency = dependency

        injected = Injected()
        assert type(injected.dependency) is DefaultConcreteDependency

    def test_abstract_base_dependency_when_resolved_with_correct_key_then_key_dependency_used(self):
        class Injected:

            @HinteDI.inject
            def __init__(self, dependency: AbstractBaseDependency, *, dependency_key: str):
                self.dependency = dependency.resolve_from_key(dependency_key)

        instance_injected = Injected(dependency_key="instance")
        default_injected = Injected(dependency_key="default")
        assert type(instance_injected.dependency) is InstanceConcreteDependency
        assert type(default_injected.dependency) is DefaultConcreteDependency

    def test_abstract_base_dependency_given_implementation_exists_then_it_can_be_used_as_dependency_alone(self):
        class Injected:

            @HinteDI.inject
            def __init__(self, dependency: InstanceConcreteDependency):
                self.dependency = dependency

        assert type(Injected().dependency) is InstanceConcreteDependency

    @pytest.mark.parametrize("dependency_type", [HinteDI.singleton_implementation, HinteDI.instance_implementation])
    def test_implementation_registration_given_no_abstract_base_then_exception_raised(self, dependency_type):
        class InvalidBase:
            pass

        class InvalidDependency:
            pass

        expected_exception = InjectionException(f"Could not register a complete implementation "
                                                f"{InvalidDependency.__name__} for class {InvalidBase.__name__} "
                                                f"because {InvalidBase.__name__} is not registered as abstract base "
                                                f"dependency.")

        with pytest.raises(InjectionException) as actual_exception:
            dependency_type(base=InvalidBase, key="foo")(InvalidDependency)
        assert str(expected_exception) == str(actual_exception.value)

    @pytest.mark.parametrize("dependency_type", [HinteDI.singleton_implementation, HinteDI.instance_implementation])
    def test_implementation_given_default_exists_when_creating_default_then_exception_thrown(self, dependency_type):
        class InvalidDefault:
            pass

        expected_exception = InjectionException(f"Could not create default implementation {InvalidDefault.__name__} "
                                                f"for base {AbstractBaseDependency.__name__} because default "
                                                f"implementation already exists as class "
                                                f"{DefaultConcreteDependency.__name__}.")

        with pytest.raises(InjectionException) as actual_exception:
            dependency_type(base=AbstractBaseDependency, key="foo", is_default=True)(InvalidDefault)
        assert str(expected_exception) == str(actual_exception.value)

    @pytest.mark.parametrize("dependency_type", [HinteDI.singleton_implementation, HinteDI.instance_implementation])
    def test_implementation_given_key_taken_when_creating_implementation_then_exception_thrown(self, dependency_type):
        class InvalidKeyName:
            pass

        expected_exception = InjectionException(f"Could not create implementation {InvalidKeyName.__name__} for base "
                                                f"{AbstractBaseDependency.__name__} because implementation with the "
                                                f"key 'default' already exists as class "
                                                f"{DefaultConcreteDependency.__name__}.")

        with pytest.raises(InjectionException) as actual_exception:
            dependency_type(base=AbstractBaseDependency, key="default")(InvalidKeyName)
        assert str(expected_exception) == str(actual_exception.value)

    @pytest.mark.parametrize("dependency_type", [HinteDI.singleton_implementation, HinteDI.instance_implementation])
    def test_implementation_when_creating_default_with_resolve_from_key_attribute_then_exception_thrown(self, dependency_type):
        @HinteDI.abstract_base
        class AnotherBase:
            pass

        class InvalidAttributeHolder:
            def resolve_from_key(self):
                pass

        expected_exception = InjectionException(f"Could not create default implementation "
                                                f"{InvalidAttributeHolder.__name__} for base {AnotherBase.__name__} "
                                                f"because default implementations cannot have an attribute named "
                                                f"'resolve_from_key'.")

        with pytest.raises(InjectionException) as actual_exception:
            dependency_type(base=AnotherBase, key="default", is_default=True)(InvalidAttributeHolder)
        assert str(expected_exception) == str(actual_exception.value)

    @pytest.mark.parametrize("dependency_type", [HinteDI.singleton_implementation, HinteDI.instance_implementation])
    def test_implementation_when_creating_default_with_resolve_from_key_attribute_then_exception_thrown(self, dependency_type):
        class ValidAttributeHolder:
            def resolve_from_key(self):
                pass

        dependency_type(base=AbstractBaseDependency, key="attribute_holder")

    def test_inject_given_base_class_with_no_implementations_when_injecting_base_then_exception_thrown(self):
        @HinteDI.abstract_base
        class UnimplementedBase:
            pass

        @HinteDI.inject
        def invalid_injection(foo: UnimplementedBase):
            pass

        expected_exception = InjectionException(f"Could not inject an implementation of abstract base class "
                                                f"{UnimplementedBase.__name__} because no implementations exist.")

        with pytest.raises(InjectionException) as actual_exception:
            invalid_injection()
        assert str(expected_exception) == str(actual_exception.value)

    def test_inject_given_base_class_with_no_default_when_injecting_then_returns_implementation_factory(self):
        @HinteDI.abstract_base
        class BaseWithNoDefault:
            pass

        @HinteDI.singleton_implementation(base=BaseWithNoDefault, key="1")
        class FirstDependency:
            pass

        class Injected:

            @HinteDI.inject
            def __init__(self, dependency: BaseWithNoDefault):
                self.dependency = dependency

        assert type(Injected().dependency) == ImplementationFactory

    def test_inject_given_implementation_factory_returned_when_calling_resolve_from_key_then_correct_dependency_returned(self):
        @HinteDI.abstract_base
        class BaseWithNoDefault:
            pass

        @HinteDI.singleton_implementation(base=BaseWithNoDefault, key="1")
        class FirstDependency:
            pass

        class Injected:

            @HinteDI.inject
            def __init__(self, dependency: BaseWithNoDefault):
                self.dependency = dependency

        assert type(Injected().dependency.resolve_from_key("1")) == FirstDependency

    def test_inject_given_implementation_factory_returned_when_calling_with_invalid_key_then_exception_thrown(self):
        @HinteDI.abstract_base
        class BaseWithNoDefault:
            pass

        @HinteDI.singleton_implementation(base=BaseWithNoDefault, key="1")
        class FirstDependency:
            pass

        class Injected:

            @HinteDI.inject
            def __init__(self, dependency: BaseWithNoDefault):
                self.dependency = dependency

        expected_exception = InjectionException(f"Could not resolve key '2' for class {BaseWithNoDefault.__name__}.")

        with pytest.raises(InjectionException) as actual_exception:
            Injected().dependency.resolve_from_key("2")
        assert str(expected_exception) == str(actual_exception.value)

    def test_inject_given_default_dependency_when_attempting_to_resolve_invalid_key_then_exception_thrown(self):
        class Injected:

            @HinteDI.inject
            def __init__(self, dependency: AbstractBaseDependency):
                self.dependency = dependency

        expected_exception = InjectionException(f"Could not resolve key '2' for class "
                                                f"{AbstractBaseDependency.__name__}.")

        with pytest.raises(InjectionException) as actual_exception:
            Injected().dependency.resolve_from_key("2")
        assert str(expected_exception) == str(actual_exception.value)

    def test_inject_given_already_resolved_then_resolve_from_key_method_not_added(self):
        class Injected:

            @HinteDI.inject
            def __init__(self, dependency: AbstractBaseDependency):
                self.dependency = dependency

        resolved_dependency = Injected().dependency.resolve_from_key("instance")
        assert not hasattr(resolved_dependency, "resolve_from_key")



from setuptools import setup, find_packages

VERSION = "0.3.0"
DESCRIPTION = "A simple dependency injector based on type hints"

with open("README.md", "r", encoding="utf-8") as readme:
    LONG_DESCRIPTION = readme.read()


setup(name="HinteDI",
      version=VERSION,
      author="Eetu Asikainen",
      author_email="eetu.asikainen1204@gmail.com",
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      packages=find_packages(),
      keywords=["dependency injection", "DI"],
      classifiers=[
          "Development Status :: 4 - Beta",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent"
      ],
      python_requires='>=3.10')
